/*
 * dx.c -- 2HD(8sec) x^C[Wt@Cwt@C𒊏o܂
 * dx.c,v 1.0 1995-04-13 18:05:16+0900 kyoko Exp
 * Copyright (C) 1995 fubuki
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include  <stdio.h>
#include  <stdlib.h>
#include  <string.h>
#include  <time.h>
#include  <utime.h>
#include  <ctype.h>
#include  <mbctype.h>
#include  <signal.h>
#include  <dirent.h>
#include  <sys/stat.h>
#include  <unistd.h>
#include  <errno.h>
#include  <doslib.h>

const char	*rcsid = "@(#)dx.c,v 1.0 1995-04-13 18:05:16+0900 kyoko Exp";
const char	*rcsname = "gainsbourg";

#define	TRUE			1
#define	FALSE			0
#define	BLOCKSIZE		1024
#define	MBLENGTH		1024
#define	MAXPATTERN		256
#define	FATTERM			0xfff
#define	FDFATMAGIC		"\xfe\xff\xff"
#define	EOS				0
#define	SLASH			"\\"

/* Micro soft disk type attribute */
#define	MS_RD			0001
#define	MS_HID			0002
#define	MS_SYS			0004
#define	MS_VOL			0010
#define	MS_DIR			0020
#define	MS_REG			0040

#define	MS_LNK			0100
#define	MS_EXE			0200

#define	FILEMASK		(S_IWUSR|S_IRUSR|S_IXUSR|S_IFDIR)

#define	ISMONTHS(m)		((m) >= 0 && (m) <= 11)
#define	ISSLASH(p)		('\\' == (p) || '/' == (p))
#define	ISPATH(p)		(ISSLASH(p) || ':' == (p))
#define	ISDRIVE(p)		(isalpha(p[0]) && ':' == (p[1]))

/* for difc */
#define	TRACK_COUNT		170
#define	SECTOR			8
#define	OFFSET_MAX		1360	/* 8sector * 153track ( + margin ) */
#define	DIFCMAGIC		"DIFC HEADER  "
#define	SEEKPTR(o)		((_seektable[(o)] * BLOCKSIZE) + _dimoffset)
enum	DIM				{ DIM_2HD_ID = 0 };

#define	PUTATTR(p)						\
{										\
	putchar(MS_DIR & p ? 'd' :			\
			MS_VOL & p ? 'v' :			\
			MS_LNK & p ? 'l' : '-');	\
	putchar(MS_REG & p ? 'a' : '-');	\
	putchar(MS_SYS & p ? 's' : '-');	\
	putchar(MS_HID & p ? 'h' : '-');	\
	putchar('r');						\
	putchar(MS_RD & p ? '-' : 'w');		\
	putchar(MS_EXE & p ? 'x' : '-');	\
}

#define	PUT_EXTMESSAGE(mode, name, c)	\
{										\
	fprintf(stderr, "%s: ", mode);		\
	if (0 != *outputdir)  {				\
		fputs(outputdir, stderr);		\
		fputs(slash, stderr);			\
	}									\
	fprintf(stderr, "%s%c", name, c);	\
}

#define	REGLEN(length)					\
({										\
	unsigned long _l;					\
	_l  = length >> 24;					\
	_l  += (length >> 8) & 0xff00;		\
	_l  += (length << 8) & 0xff0000;	\
	_l  += (length << 24) & 0xff000000;	\
	(_l);								\
})

typedef struct {
	char			name[8];
	char			ext[3];
	char			atr;
	char			extname[10];
	unsigned short	time;
	unsigned short	date;
	unsigned short	start_fat;
	unsigned long	length;
} DIRENT;

void		putdir(DIRENT *pool);
int			exdir(DIRENT *pool, char *path, short **reg, int count);
int			getfatlink(int fatno, char *fat);
void		readdirs(int fat, char *path, short **reg, int count);
int			chkarg(int argc, char **argv, char **file);
void		usage(void);
int			writefile(DIRENT *pool, char *name);
time_t		maketime(DIRENT *pool);
char		*fre_compile(short *dbuf, int dsize, const char *string);
void		xfrec(short *dbuf, int dsize, const char *string);
int			fre_match(const short *regex, const char *string);
void		*xmalloc(size_t foo);
void		regtime(struct tm *tp, DIRENT *pool);
int			xfseek(FILE *fp, long offset, int whence);
int			xfread(void *buffer, size_t size, size_t so, FILE *fp);
int			xfwrite(void *buffer, size_t size, size_t so, FILE *fp);
void		xfclose(FILE *fp);
char		*getname(char *p);
void		xmkdir(char *buffer, char *last);
void		signalhandler(int sig);
void		xutime(char *name, DIRENT *pool);
int			xexit(int n);
char		*getpwd(void);
char		*exist_interactive(char *name);
void		ifdim2hd(int a, char *name);
void		make_seek_table(void);
int			eximatch(char *file, char *ext);
char		*chopslash(char *name);
void		xchdir(char *path);

static const char	*months[] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec",
	"---"
};

extern	int ignore_case;

char	*This;
char	*Fat;
FILE	*fp;
int		indent = 0;
short	*reg[MAXPATTERN];
char	*slash;
char	newname[128];
char	*wfile;
char	*_tracktable;
short	*_seektable;
int		_dimoffset = 0;

char	*outputdir = "";
char	*pwd;

int		iopt = FALSE;
int		jopt = FALSE;
int		xopt = FALSE;
int		lopt = FALSE;
int		fopt = FALSE;
int		qopt = FALSE;
int		vopt = FALSE;
int		oopt = FALSE;
int		popt = FALSE;
int		nopt = FALSE;
int		once = FALSE;
int		copt = FALSE;
int		Dopt = FALSE;

int
main(int argc, char **argv)
{
	DIRENT	*pool = xmalloc(sizeof(DIRENT) * 192);
	char	*file[MAXPATTERN];
	char	imagefile[256];
	int		count;

	slash = getenv("SLASH") ?: SLASH;
	pwd   = getpwd();
	This  = getname(argv[0]);
	Fat   = xmalloc(BLOCKSIZE * 2);

	if (0 == (count = chkarg(--argc, ++argv, file))) {
		usage();
		/* NOTREACHED */
	}
	if (iopt) ignore_case = TRUE;
	count--;
	if (0 == count) {
		count++;
		file[1] = "*";
		lopt = TRUE;
	}
	if (EOS != *outputdir) {
		if (!ISDRIVE(file[0]) && !ISSLASH(*file[0])) {
			strcpy(imagefile, pwd ?: "");
			strcat(imagefile, slash);
			strcat(imagefile, file[0]);
		} else
			strcpy(imagefile, file[0]);
		xchdir(outputdir);
	} else
		strcpy(imagefile, file[0]);

	if (NULL == (fp = fopen(imagefile, "rb"))) {
		fprintf(stderr, "%s: file can't open: %s\n", This, imagefile);
		xexit(1);
		/* NOTREACHED */
	}
	signal(SIGINT, signalhandler);
	if (Dopt) {
		ifdim2hd(fgetc(fp), imagefile);
		_tracktable = xmalloc(TRACK_COUNT);
		xfread(_tracktable, sizeof(char), TRACK_COUNT, fp);
		xfread(Fat, sizeof(char), 255 - TRACK_COUNT, fp);	/* SKIP DIM header */
		if (0 != strncmp(Fat, DIFCMAGIC, 13)) {
			ifdim2hd(!DIM_2HD_ID, imagefile);
			/* NOTREACHED */
		}
	}
	xfseek(fp, BLOCKSIZE, SEEK_CUR);						/* SKIP IPL */
	xfread(Fat, sizeof(char), BLOCKSIZE * 2, fp);			/* read FAT */
	if (!fopt && 0 != strncmp(Fat, FDFATMAGIC, 3)) {
		fprintf(stderr, "%s: this file not HDM format: %s\n", This, file[0]);
		xfclose(fp);
		xexit(1);
		/* NOTREACHED */
	}
	xfseek(fp, BLOCKSIZE * 2, SEEK_CUR);					/* SKIP FAT2 */
	xfread(pool, sizeof(char), sizeof(DIRENT) * 192, fp);	/* read Directory Entry */
	make_seek_table();
	{
		int i;
		for(i = 0; i < count; i++) {
			reg[i] = (short *)xmalloc(sizeof(short) * MBLENGTH);
			xfrec(reg[i], MBLENGTH, file[i+1]);
		}
	}
	exdir(pool, "", reg, count);
	xfclose(fp);
	/* free(pool); free(Fat); free(_seektable); loop(n) free(reg[n]); */
	return xexit(0);
}

static void
usage(void)
{
	/*
		ŕ\ȂIvVgpꍇ
		x DOS ɑ΂mȂƁAn}ɂƂł傤B
		˂Ďgpꍇ͊e̐ӔCɉĎgpĂB
	*/
	fprintf(stderr, "usage: %s [-l1pxijcqnoV] [-d outputdir] imagefile [-AD] [extfile] ...\n", This);
	xexit(1);
}

static int
chkarg(int argc, char **argv, char **file)
{
	int		i = 0;

	while(argc--) {
		if ('-' == **argv) {
			while(*++*argv) {
				switch(**argv) {

				case	'V':	/* version put */
					printf("%s: version %s (C) 1995 fubuki\n", This, rcsname);
					xexit(0);
					/* NOTREACHED */
				case	'i':	/* p^[ő啶 */
					iopt = TRUE;
					break;
				case	'q':	/* ܂ */
					qopt = TRUE;
					break;
				case	'j':	/* Xn (mkdir Ȃ) */
					jopt = TRUE;
					xopt = TRUE;
					break;
				case	'f':	/* ۏ */
					fopt = TRUE;
					break;
				case	'p':	/* stdout  o */
					popt = TRUE;
					xopt = TRUE;
					jopt = TRUE;
					break;
				case	'c':	/* Human g file name(10)؂̂ */
					copt = TRUE;
					break;
				case	'v':	/* volume o */
					vopt = TRUE;
					break;
				case	'x':	/* o */
					xopt = TRUE;
					break;
				case	'D':	/* *.dim mode (2HD(8sec) only) */
					Dopt = TRUE;
					break;
				case	'A':	/* all over (x^ mode (2HD(8sec) only)) */
					Dopt = FALSE;
					break;
				case	'o':	/*  over write */
					oopt = TRUE;
					break;
				case	'n':	/*  skip */
					nopt = TRUE;
					break;
				case	'1':	/* Xg */
					once = TRUE;
					break;
				case	'l':	/* Xg (l) */
					lopt = TRUE;
					break;
				case	'd':	/* o path w */
					if (EOS == *++*argv) {
						++argv;
						if (0 > --argc) usage();
						/* NOTREACHED */
					}
						outputdir = chopslash(*argv);
						goto DOWN;
				default:
					usage();
					/* NOTREACHED */
				}
			}
		} else {
			if (eximatch(*argv, "dim")) Dopt = TRUE;
			file[i++] = *argv;
			if (i > MAXPATTERN) {
				fprintf(stderr, "%s: too many pattern\n", This);
				xexit(1);
				/* NOTREACHED */
			}
		}
		DOWN:
		++argv;
	}
	return i;
}

/* A͂Ȃ!! */
static void
xmkdir(char *buffer, char *last)
{
	char		line[64], *p, *q;

	p = buffer;
	q = line;
	while(*p) {
		if (ismbblead(*p)) {
			p += 2;
		} else if (p > last && *p == *slash) {
			*q = EOS;
			if (0 == mkdir(line, FILEMASK)) {
				if (!qopt) PUT_EXTMESSAGE("mkdir", line, '\n');
			}
			xmkdir(buffer, p);
			return;
		} else {
			*q++ = *p++;
		}
	}
}

#ifndef UTIME_EXIST
static void
xutime(char *name, DIRENT *pool)
{
	FILE	*p;
	unsigned int	tstamp, a;

	a = _dos_chmod(name, -1);
	_dos_chmod(name, MS_REG);
	tstamp = ((pool->date << 8) | (pool->date >> 8)) & 0xffff;
	tstamp <<= 16;
	tstamp |= ((pool->time << 8) | (pool->time >> 8)) & 0xffff;
	p = fopen(name, "a");
	_dos_filedate(fileno(p), tstamp);
	fclose(p);
	_dos_chmod(name, a);
}
#endif /* UTIME_EXIST */

/*
	т̖{(т)
 */
static int
exdir(DIRENT *pool, char *path, short **reg, int count)
{
	char	buffer[256];
	char	*p, *q, *pathterm;
	int		i, cond;

	strcpy(buffer, path);
	if (EOS != *path) strcat(buffer, slash);
	pathterm = &buffer[strlen(buffer)];
	while(0 != pool->name[0]) {
		if ('\xe5' == pool->name[0]) {
			pool++;
			continue;
		}
		if (0 == (strncmp(pool->name, ". ", 2)) ||
			0 == (strncmp(pool->name, ".. ", 3))) {
			pool++;
			continue;
		}
		q = pathterm;
		p = pool->name;
		if (5 == *p) *q = 0xe5;
		else *q = *p;
		p++;
		q++;
		i = 7;
		while(i-- && (' ' != *p)) *q++ = *p++;
		if (!copt) {
			p = pool->extname;
			i = 10;
			while(i-- && (0 != *p)) *q++ = *p++;
		}
		p = pool->ext;
		if (' ' != *p) {
			*q++ = '.';
			i = 3;
			while(i-- && (' ' != *p)) *q++ = *p++;
		}
		*q = EOS;
		for(i = 0; i < count; i++)
		if (fre_match(reg[i], buffer)) {
			*newname = 0;
			cond = FALSE;
			if (xopt) {
				if (vopt && (MS_VOL & pool->atr)) {
					FILE	*p;
					if (!qopt) PUT_EXTMESSAGE("extract volume label", buffer, '\n');
					_dos_chmod(buffer, MS_REG);
					p = fopen(buffer, "w");
					xfclose(p);
					cond = TRUE;
				} else if (MS_DIR & pool->atr) {
					int		n;
					if (!jopt && indent) xmkdir(buffer, buffer);
					n = mkdir(jopt ? getname(buffer) : buffer, FILEMASK);
					if (!qopt && (0 == n))
						PUT_EXTMESSAGE("extract directory", buffer, '\n');
					cond = TRUE;
				} else if (!(MS_VOL & pool->atr)) {
					if (!jopt && indent) xmkdir(buffer, buffer);
					cond = writefile(pool, jopt ? getname(buffer) : buffer);
				}
				if (cond && !popt) {
#if UTIME_EXIST
					struct utimbuf tm;
					tm.modtime = tm.actime = maketime(pool);
					utime(0 != *newname ? newname :
							jopt ? getname(buffer) : buffer, &tm);
					/*
						libc  utime  [N21?
						̂ŁAƂ肠ȉœ
					*/
#else	/* Human Local */
					xutime(0 != *newname ? newname :
							jopt ? getname(buffer) : buffer, pool);
#endif
					_dos_chmod(0 != *newname ? newname :
							jopt ? getname(buffer) : buffer, pool->atr);
				}
			} else {
				if (!once) {
					struct tm tp;
					PUTATTR(pool->atr);
					printf(" %7d ", REGLEN(pool->length)); /* FD  1.2M Ȃ̂ %7d */
					regtime(&tp, pool);
					printf("%s %2d %02d:%02d:%02d %4d ",
						months[tp.tm_mon], tp.tm_mday,
						tp.tm_hour, tp.tm_min, tp.tm_sec,
						tp.tm_year + 1900
					);
				}
				fputs(buffer, stdout);
				if (MS_DIR & pool->atr) putchar(*slash);
				putchar('\n');
			}
		} /* for loop end */
		if (MS_DIR & pool->atr) {
			unsigned short f = pool->start_fat >> 8 | pool->start_fat << 8;
			indent ++;
			readdirs(f, buffer, reg, count);
				/* ̊֐̒Ō݂̊֐ċAI call Ă */
			indent --;
		}
		pool++;
	}
	return 0;
}

static char *
exist_interactive(char *name)
{
	char	buffer[32];
	FILE	*fp;

	fp = fopen(name, "r");
	if (NULL != fp) {
		fclose(fp);
		if (nopt) return NULL;
		fprintf(stderr, "writable %s exists; remove it? [nyr](n):", name);
		gets(buffer);
		switch(*buffer) {

		case	'r':
			fprintf(stderr, "new name: %s -> ", name);
			gets(newname);
			name = exist_interactive(newname);
			break;

		case	'y':
#if FORCE_WRITE
			if ((MS_RD|MS_HID|MS_SYS) & _dos_chmod(name, -1)) {
				_dos_chmod(name, MS_REG);
				_dos_delete(name);
			}
#endif /* FORCE_WRITE */
			break;

		default:
			 return NULL; 		/* "checkout aborted" */

		}
	}
	return name;
}

static int
writefile(DIRENT *pool, char *name)
{
	FILE			*fpw;
	unsigned short	fat = pool->start_fat >> 8 | pool->start_fat << 8;
	char			buffer[BLOCKSIZE];
	unsigned long	l, s, n = 0;

	if (!oopt && !popt) {
		if (NULL == (name = exist_interactive(name))) return FALSE;
	}
	if (!qopt) PUT_EXTMESSAGE("extract", name, ' ');
	if (popt) {
		fpw = stdout;
	} else {
		if (NULL == (fpw = fopen(name, "wb"))) {
			fprintf(stderr, "%s: file can't write: %s\n", This, name);
			xexit(1);
			/* NOTREACHED */
		}
		wfile = name;
			/* ݏ݃I[vĂt@C signalpɕێ */
	}
	xfseek(fp, SEEKPTR(11 + (fat - 2)), SEEK_SET);
	l = REGLEN(pool->length);
	s = BLOCKSIZE;
	if (l < BLOCKSIZE) s = l;
	xfread(buffer, sizeof(char), s, fp);
	xfwrite(buffer, sizeof(char), s, fpw);
	l -= s;
	while(FATTERM != (fat = getfatlink(fat, Fat))) {
		xfseek(fp, SEEKPTR(11 + (fat - 2)), SEEK_SET);
		if (l < BLOCKSIZE) s = l;
		xfread(buffer, sizeof(char), s, fp);
		xfwrite(buffer, sizeof(char), s, fpw);
		if (!qopt && (0 == n % 16)) {
			fputc('.', stderr);
			fflush(stdout);
		}
		l -= s;
		n++;
	}
	if (!qopt) fputc('\n', stderr);
	if (!popt) xfclose(fpw);
	return TRUE;
}

static void
readdirs(int fat, char *path, short **reg, int count)
{
	int		fats = fat;
	int		n = 1;
	char	*sub, *p;

	while(FATTERM != (fats = getfatlink(fats, Fat))) n++;
	p = sub = xmalloc(BLOCKSIZE * n);
	xfseek(fp, SEEKPTR(11 + (fat - 2)), SEEK_SET);
	xfread(p, sizeof(char), BLOCKSIZE, fp);
	p += BLOCKSIZE;
	while(FATTERM != (fats = getfatlink(fat, Fat))) {
		xfseek(fp, SEEKPTR(11 + (fats - 2)), SEEK_SET);
		xfread(p, sizeof(char), BLOCKSIZE, fp);
		p += BLOCKSIZE;
		fat = fats;
	}
	exdir((DIRENT *)sub, path, reg, count);
	free(sub);
}

/* 12bit FAT only */
static int
getfatlink(int fatno, char *fat)
{
	unsigned char *p = (unsigned char *)fat;
	int		offset, off, link;

	offset = fatno * 15;
	off = offset / 10;
	if (5 == offset % 10) {
		link =  *(p + off + 1) << 4;
		link += ((*(p + off) >> 4) & 017);
	} else {
		link =  (*(p + off + 1) & 017 ) << 8;
		link += *(p + off);
	}
	return link;
}

static void
regtime(struct tm *tp, DIRENT *pool)
{
	tp->tm_sec = ((pool->time >> 8) & 037) * 2;
	tp->tm_min = (pool->time >> 13) | ((pool->time << 3) & 070);
	tp->tm_hour = (pool->time >> 3 ) & 037;
	tp->tm_mday = (pool->date >> 8) & 037;
	tp->tm_mon = (((pool->date >> 13) & 07) | ((pool->date << 3) & 010)) - 1;
	tp->tm_year = ((pool->date >> 1) & 0177) + 80;
	tp->tm_isdst = -1;

	if (!ISMONTHS(tp->tm_mon)) tp->tm_mon = 12;
}

#if UTIME_EXIST
static time_t
maketime(DIRENT *pool)
{
	struct tm n;

	regtime(&n, pool);
	return mktime(&n);
}
#endif /* UTIME_EXIST */

static void *
xmalloc(size_t foo)
{
	void	*p;
	if (NULL == (p = malloc(foo))) {
		fprintf(stderr, "%s: memory exhausted\n", This);
		xexit(1);
		/* NOTREACHED */
	}
	return p;
}

static int
xfseek(FILE *fp, long offset, int whence)
{
	int		n;
	if (0 > (n = fseek(fp, offset, whence))) {
		fprintf(stderr, "%s: file seek error\n", This);
		xexit(1);
		/* NOTREACHED */
	}
	return n;
}

static int
xfread(void *buffer, size_t so, size_t size, FILE *fp)
{
	int		n;
	n = fread(buffer, so, size, fp);
	if ((n < (so * size)) || (0 != ferror(fp))) {
		fprintf(stderr, "%s: file read error\n", This);
		fclose(fp);
		xexit(1);
		/* NOTREACHED */
	}
	return n;
}

static int
xfwrite(void *buffer, size_t so, size_t size, FILE *fp)
{
	int		n;
	n = fwrite(buffer, so, size, fp);
	if ((n < (so * size)) || (0 != ferror(fp))) {
		fprintf(stderr, "%s: file write error\n", This);
		fclose(fp);
		xexit(1);
		/* NOTREACHED */
	}
	return n;
}

static void
xfclose(FILE *fp)
{
	if (0 != fclose(fp)) {
		fprintf(stderr, "%s: file close error\n", This);
		xexit(1);
		/* NOTREACHED */
	}
}

void
xfrec(short *dbuf, int dsize, const char *string)
{
	char	*p;
	if (NULL != (p = fre_compile(dbuf, dsize, string))) {
		fprintf(stderr, "%s: %s\n", This, p);
		xexit(1);
		/* NOTREACHED */
	}
}

void
xchdir(char *path)
{
	extern	errno;

	if (-1 == chdir(path)) {
		switch(errno) {

		case	ENOENT:
			fprintf(stderr, "%s: directory not found: %s\n", This, path);
			break;
		case	ENOTDIR:
			fprintf(stderr, "%s: bad directory name: %s\n", This, path);
			break;
		case	ELOOP:
			fprintf(stderr, "%s: deep nest symbolic link: %s\n", This, path);
			break;

		}
		xexit(1);
		/* NOTREACHED */
	}
	return;
}

void
make_seek_table(void)
{
	short	*p = _seektable = xmalloc(sizeof(short) * OFFSET_MAX);
	char	*t = _tracktable;
	int		i, j, k;

	if (Dopt) {
		_dimoffset = 256;
		for(k = i = 0; i < TRACK_COUNT; i++, t++) {
			for(j = 0; j < SECTOR; j++) {
				*p++ = *t ? k++ : k ;
			}
		}
		free(_tracktable);	/* Ȃ */
	} else {
		for(i = 0; i < OFFSET_MAX; i++) *p++ = i;
	}
}

static void
ifdim2hd(int a, char *name)
{
	if (DIM_2HD_ID != a) {
		fprintf(stderr, "%s: this file not 2HD dim format: %s\n", This, name);
		xfclose(fp);
		xexit(1);
		/* NOTREACHED */
	}
}

char *
chopslash(char *name)
{
	char	*p = name;
	while(*p) {
		if (ismbblead(*p)) {
			p += 2;
			continue;
		}
		if (1 == strlen(p)) {
			if (ISSLASH(*p)) {
				*p = EOS;
				break;
			}
		}
		p++;
	}
	return name;
}

static int
eximatch(char *file, char *ext)
{
	char	*q, *p = file;
	int		r;

	q = p;
	while(*p) {
		if (ismbblead(*p)) {
			p += 2;
			continue;
		}
		if ('.' == *p) q = p+1;
		p++;
	}
	if (3 != strlen(q)) r = FALSE;
	if (0 == stricmp(q, ext)) r = TRUE;
	else r = FALSE;
	return r;
}

static char *
getpwd(void)
{
#if NO_PWD_ENV
	static char	p[_MAX_PATH], *q;
	q = p;
	*q++ = _dos_curdrv() + 'A';
	*q++ = ':';
	*q++ = *slash;
	_dos_curdir(0, q);
	q = p;
	while(*q) {
		if (ismbblead(*q)) {
			q += 2;
			continue;
		}
		if (ISSLASH(*q)) *q = *slash;
		q++;
	}
	*q = EOS;
	return p;
#else
	return getenv("PWD");
#endif
}

static char *
getname(char *p)
{
	char	*q = p;

	while(*p) {
		if (ismbblead(*p)) {
			p += 2;
			continue;
		}
		if (ISPATH(*p)) q = p+1;
		p++;
	}
	return q;
}

static void
signalhandler(int sig)
{
	fcloseall();
	if (!popt) _dos_delete(wfile);
	_dos_print("aborted\r\n");
	xexit(2);
}

static int
xexit(int n)
{
	if (EOS != *outputdir) chdir(pwd);
	exit(n);
	return 0;	/* for `main' exit dummy */
}

/*** dx.c ***/
